#if !defined(CREATURES_BRAIN_H)
#define CREATURES_BRAIN_H
//-------------------------------------------------------------------
// Various classes and data structures to obtain information on a
// what is going on inside the brain for a Norn.
//
// Compiled and tested under Visual C++ v4.2.
//
// Some notes:
//   - Almost zero error handling.
//
// Provided by John Booth and in the public domain. Use as you wish.
// jwbooth@mailcity.com
//-------------------------------------------------------------------
#include <memory>
#include <vector>
#include <string>

#pragma warning (disable: 4786)

struct LobeData
{
	LobeData();
	LobeData(unsigned int cell,
		unsigned int output, 
		unsigned int state,
		unsigned int dendrites,
		unsigned int susceptability,
		unsigned int stw,
		unsigned int ltw,
		unsigned int strength);

	enum
	{
		kOutput,
		kState,
		kNumberOfDendrites,
		kTotalSusceptability,
		kTotalSTW,
		kTotalLTW,
		kTotalStrength,
		kLast
	};

	int getOutput() const { return fData[kOutput]; }
	int getState() const { return fData[kState]; }
	int getNumberOfDendrites() const { return fData[kNumberOfDendrites]; }
	int getTotalSusceptability() const { return fData[kTotalSusceptability]; }
	int getTotalSTW() const { return fData[kTotalSTW]; }
	int getTotalLTW() const { return fData[kTotalLTW]; }
	int getTotalStrength() const { return fData[kTotalStrength]; }
	int getCell() const { return fCell; }

	unsigned int fCell;
	unsigned int fData[kLast];

	int operator<(LobeData const& rhs) const
	{
		return getOutput() < rhs.getOutput();
	}

	int operator==(LobeData const& rhs) const
	{
		return getOutput() == rhs.getOutput();
	}
};


int outputIsZero(LobeData rhs);
int stateIsZero(LobeData rhs);
int outputIsGreater(LobeData const& lhs, LobeData const& rhs);
int stateIsGreater(LobeData const& lhs, LobeData const& rhs);

struct Lobe
{
	typedef std::vector<LobeData> CellVector;

	virtual ~Lobe() {}
	virtual int getLobeNumber() const = 0;
	virtual std::string getMacroCommand() const;
	virtual std::string getCellDescription(int cell) const = 0;
	virtual unsigned int getNumberOfCells() const = 0;
	CellVector parseCellData(std::string const& result) const;
};

struct PerceptionLobe : Lobe
{
	virtual int getLobeNumber() const;
	virtual std::string getCellDescription(int cell) const;
	virtual unsigned int getNumberOfCells() const;
};

struct ConceptLobe : Lobe
{
	virtual int getLobeNumber() const;
	virtual std::string getCellDescription(int cell) const;
	virtual unsigned int getNumberOfCells() const;
};

struct SensoryLobe : Lobe
{
	virtual int getLobeNumber() const;
	virtual std::string getCellDescription(int cell) const;
	virtual unsigned int getNumberOfCells() const;
};

struct DecisionLobe : Lobe
{
	virtual int getLobeNumber() const;
	virtual std::string getCellDescription(int cell) const;
	virtual unsigned int getNumberOfCells() const;
};

struct AttentionLobe : Lobe
{
	virtual int getLobeNumber() const;
	virtual std::string getCellDescription(int cell) const;
	virtual unsigned int getNumberOfCells() const;
};

struct NounLobe : AttentionLobe
{
	virtual int getLobeNumber() const;
};

struct VerbLobe : DecisionLobe
{
	virtual int getLobeNumber() const;
};


struct StimulusSourceLobe : NounLobe
{
	virtual int getLobeNumber() const;
	virtual unsigned int getNumberOfCells() const;
};



struct DriveLobe : Lobe
{
	virtual int getLobeNumber() const;
	virtual std::string getCellDescription(int cell) const;
	virtual unsigned int getNumberOfCells() const;
};

// A composite lobe - one that allows adding/removing
// of cells from other lobes and manipulating that data
// as if it was an independant lobe.
struct CompositeLobe : Lobe
{
	struct Item
	{
		Item() : fLobe(0), fCell(0), fDendrite(0) {}
		Item(unsigned int lobe, unsigned int cell, unsigned int dendrite, std::string const& desc = std::string("")) :
			fLobe(lobe),
			fCell(cell),
			fDendrite(dendrite),
			fDescription(desc)
		{
		}
		unsigned int fLobe;
		unsigned int fCell;
		unsigned int fDendrite;
		std::string fDescription;
		int operator==(Item const& rhs) const
		{
			return 
				fLobe == rhs.fLobe && 
				fCell == rhs.fCell && 
				fDendrite == rhs.fDendrite;
		}
		int operator!=(Item const& rhs) const
		{
			return !operator==(rhs);
		}
	};

	typedef std::vector<Item> ItemList;
	ItemList fItems;
	void addCell(unsigned int lobe, unsigned int cell, unsigned int dendrite, std::string const& desc);
	void removeCell(unsigned int index);
	virtual int getLobeNumber() const { return 999; }
	virtual std::string getMacroCommand() const;
	virtual std::string getCellDescription(int cell) const;
	virtual unsigned int getNumberOfCells() const;
};

struct Brain
{
	struct BrainCell
	{
		BrainCell() : fLobeName("Unknown"), fCellName("Unknown") {}
		BrainCell(std::string const& lobeName, std::string const& cellName) : 
			fLobeName(lobeName), fCellName(cellName) {}
		std::string fLobeName;
		std::string fCellName;
	};

	BrainCell fCells[64][48];

	Brain();
	std::string getLobeName(unsigned int x, unsigned int y);
	std::string getCellDescription(unsigned int x, unsigned int y);
	struct Results
	{
		Results() : fX(0), fY(0), fValue(0) {}
		Results(int x, int y, int v) : fX(x), fY(y), fValue(v) {}
		unsigned int fX;
		unsigned int fY;
		unsigned int fValue;
	};
	typedef std::vector<Results> ResultsVector;

	ResultsVector parseFromActivity(std::string const& activity);
};

#endif
